@@ -20,12 +20,13 @@
  *
  */
 
-/* $Id: execute.c,v 1.45 2006-09-07 17:15:13 timo Exp $ */
+/* $Id: execute.c,v 1.67 2007-05-30 07:32:48 timo Exp $ */
 
 #include "ffe.h"
 #include <stdlib.h>
 #include <string.h>
 #include <ctype.h>
+#include <stdio.h>
 
 #ifdef PACKAGE
 static char *program = PACKAGE;
@@ -39,6 +40,8 @@ static char *program = "ffe";
 #define FIELD_SIZE (8 * 1024)
 #define WRITE_BUFFER (512 * 1024)
 
+extern struct replace *replace;
+
 static struct input_file *files = NULL;
 static struct input_file *current_file = NULL;
 static FILE *input_fp = NULL;
@@ -103,7 +106,7 @@ close_output_file()
 void
 set_input_file(char *name)
 {
-    struct input_file *f = files;
+    register struct input_file *f = files;
     
     if(files == NULL)
     {
@@ -230,7 +233,7 @@ get_field_count(char quote, char *type, char *line)
         {
             inside_quote = !inside_quote;
         }
-        p++;
+        if(*p) p++;
     }
     return fields;
 }
@@ -240,7 +243,7 @@ get_field_count(char quote, char *type, char *line)
 char *
 get_fixed_field(int position, int length, int line_length, char *line)
 {
-    char *t,*s;
+    register char *t,*s;
 
     if(length >= field_buffer_size - 1) 
     {
@@ -272,10 +275,10 @@ get_fixed_field(int position, int length, int line_length, char *line)
 char *
 get_separated_field(int position, char quote, char *type, char *line)
 {
-    char *p = line;
+    register char *p = line;
     int fieldno = 1;
     int inside_quote = 0;
-    int i = 0;
+    register int i = 0;
 
     while(*p && fieldno <= position)
     {
@@ -315,7 +318,7 @@ get_separated_field(int position, char quote, char *type, char *line)
                 field_buffer = xrealloc(field_buffer,field_buffer_size);
             }
         }
-        p++;
+        if(*p) p++;
     }
     field_buffer[i] = 0;
     return field_buffer;
@@ -387,11 +390,14 @@ vote(int bindex)
     {
         r = s->r;
         votes = 0;
-        while(r != NULL && !votes)
+        if(s->vote == bindex)   // check only structures having all records matched so far
         {
-            votes = vote_record(s->quote,s->type,s->header,r,guess_line_length[bindex],guess_buffer[bindex]);
-            s->vote += votes;             /* only one vote per line */
-            r = r->next;
+            while(r != NULL && !votes)
+            {
+                votes = vote_record(s->quote,s->type,s->header,r,guess_line_length[bindex],guess_buffer[bindex]);
+                s->vote += votes;             /* only one vote per line */
+                r = r->next;
+            }
         }
         total_votes += votes;
         s = s->next;
@@ -399,7 +405,7 @@ vote(int bindex)
     if(!total_votes && no_matching_lines < NO_MATCH_LINES)
     {
         no_matching_lines++;
-        fprintf(stderr,"%s: Line %ld in \'%s\' does not match any structure\n",program,current_file->lineno,current_file->name);
+        fprintf(stderr,"%s: Line %ld in \'%s\' does not match, line length = %d\n",program,current_file->lineno,current_file->name,guess_line_length[bindex]);
     }
 }
 
@@ -492,15 +498,18 @@ writec(char c)
 
 
 /* write string to write buffer */
-void
+inline void
 writes(char *string)
 {
     register char *s = string;
 
-    while(*s)
+    if(s != NULL)
     {
-        writec(*s);
-        s++;
+        while(*s)
+        {
+            writec(*s);
+            s++;
+        }
     }
 }
 
@@ -576,12 +585,17 @@ print_text(struct structure *s, struct record *r,char *buffer)
 
 /* prints a header text */
 /* text can contain only one %-directive %n */
-void
+/* return values: */
+/* 1 - Header is printed or header should not be printed */
+/* 0 - Header is not printed and should be printed */
+int
 print_header(struct structure *s, struct record *r)
 {
     struct print_field *pf = r->pf;
 
-    if(r->o == no_output || r->o->header == NULL) return;
+    if(r->o == no_output || r->o->header == NULL) return 1; /* no header for this run */
+
+    if(pf == NULL) return 0;   /* no printable fields for this record */
 
     start_write();
 
@@ -614,6 +628,8 @@ print_header(struct structure *s, struct record *r)
     }
     writes(r->o->record_trailer);
     flush_write();
+
+    return 1;  /* header printed */
 }
 
 /* returns pointer to next input line
@@ -732,7 +748,7 @@ update_field_positions(char *type,char quote,struct field *fields,int len,char *
                 {
                     inside_quote = !inside_quote;
                 }
-                p++;
+                if(*p) p++;
             }
             break;
     }
@@ -759,19 +775,28 @@ select_record(struct structure *s,int length,char *buffer)
 void
 print_fixed_field(char format,struct field *f,char *buffer)
 {
-    int i = 0;
+    register int i = 0;
+    register char *data;
     char *start = write_pos;
 
+    if(f->const_data != NULL)
+    {
+        data = f->const_data;
+    } else
+    {
+        data = &buffer[f->bposition];
+    }
+
     switch(format)
     {
         case 'd':
         case 'D':
         case 'e':
-            while(i < f->length) writec(buffer[f->bposition + i++]);
+            while(i < f->length) writec(data[i++]);
             break;
         case 't':
-            while(isspace(buffer[f->bposition + i])) i++;
-            while(i < f->length) writec(buffer[f->bposition + i++]);
+            while(isspace(data[i])) i++;
+            while(i < f->length) writec(data[i++]);
             if(write_pos > start && isspace(write_pos[-1]))
             {
                 write_pos--;
@@ -782,7 +807,7 @@ print_fixed_field(char format,struct field *f,char *buffer)
     }
 }
 
-/* print a single seprated field */
+/* print a single separated field */
 void
 print_separated_field(char format,char quote,char separator,struct field *f,char *buffer)
 {
@@ -792,60 +817,122 @@ print_separated_field(char format,char quote,char separator,struct field *f,char
 
     start = write_pos;
 
-    if(f->bposition < 0)
+    if(f->const_data != NULL)   // use fixed field printing for const data
     {
-        if(format == 'D') while(write_pos - start < f->length) writec(' ');
-        return;
-    }
-
-
-    switch(format)
+        print_fixed_field(format,f,buffer);
+    } else
     {
-        case 'd':
-        case 't':
-        case 'D':
-        case 'e':
-            p = &buffer[f->bposition];
-            if(quote || format == 't') while(isspace(*p)) p++;
-            if(*p == quote && quote) 
-            {
-                p++;
-                inside_quote = 1;
-                if(format == 't') while(isspace(*p)) p++;
-            }
-            while((*p != separator || inside_quote) && *p)
-            {
-                if(((*p == quote && p[1] == quote) || (*p == '\\' && p[1] == quote)) && quote) 
+        if(f->bposition < 0)
+        {
+            if(format == 'D') while(write_pos - start < f->length) writec(' ');
+            return;
+        }
+
+        switch(format)
+        {
+            case 'd':
+            case 't':
+            case 'D':
+            case 'e':
+                p = &buffer[f->bposition];
+                if(quote || format == 't') while(isspace(*p)) p++;
+                if(*p == quote && quote) 
                 {
                     p++;
-                } else if(*p == quote)
-                {
-                    if(inside_quote)
-                    {
-                        if(format == 't' && isspace(write_pos[-1]))
-                        {
-                            write_pos--;     
-                            while(write_pos > start && isspace(*write_pos)) write_pos--;
-                            write_pos++;
-                        }
-                        if(format == 'D') while(write_pos - start < f->length) writec(' ');
-                        return;
-                    }
+                    inside_quote = 1;
+                    if(format == 't') while(isspace(*p)) p++;
                 }
-                writec(*p);
-                p++;
-            }            
-            if(format == 't' && isspace(write_pos[-1]))
-            {
-                write_pos--;     
-                while(write_pos > start && isspace(*write_pos)) write_pos--;
-                write_pos++;
-            }
-            if(format == 'D') while(write_pos - start < f->length) writec(' ');
-            break;
+                while((*p != separator || inside_quote) && *p)
+                {
+                    if(((*p == quote && p[1] == quote) || (*p == '\\' && p[1] == quote)) && quote) 
+                    {
+                        p++;
+                    } else if(*p == quote)
+                    {
+                        if(inside_quote)
+                        {
+                            if(format == 't' && isspace(write_pos[-1]))
+                            {
+                                write_pos--;     
+                                while(write_pos > start && isspace(*write_pos)) write_pos--;
+                                write_pos++;
+                            }
+                            if(format == 'D') while(write_pos - start < f->length) writec(' ');
+                            return;
+                        }
+                    }
+                    writec(*p);
+                    if(*p) p++;
+                }            
+                if(format == 't' && isspace(write_pos[-1]))
+                {
+                    write_pos--;     
+                    while(write_pos > start && isspace(*write_pos)) write_pos--;
+                    write_pos++;
+                }
+                if(format == 'D') while(write_pos - start < f->length) writec(' ');
+                break;
+        }
     }
 }
 
+/* search the lookup table, return the found value */
+char *
+make_lookup(struct lookup *l,char *search)
+{
+    register char *k,*s;
+    int search_len = strlen(search);
+    int max_len = -1;
+    char *ret_val = NULL;
+    char *longest = NULL;
+    struct lookup_data *d = l->data;
+
+    switch(l->type)
+    {
+        case EXACT:
+            if(search_len <= l->max_key_len)
+            {
+                while(d != NULL && ret_val == NULL)
+                {
+                    k = d->key;
+                    s = search;
+                    while(*k == *s && *k) 
+                    {
+                        k++;
+                        s++;
+                    }
+                    if(!*s && !*k) ret_val = d->value;
+                    d = d->next;
+                }
+            }
+            break;
+        case LONGEST:
+            while(d != NULL && max_len < l->max_key_len)
+            {
+                k = d->key;
+                s = search;
+                while(*k == *s && *k) 
+                {
+                    k++;
+                    s++;
+                }
+                if(!*k && d->key_len > max_len)
+                {
+                    max_len = d->key_len;
+                    longest = d->value;
+                }
+                d = d->next;
+            }
+            ret_val = longest; 
+            break;
+    }
+
+    if(ret_val == NULL) ret_val = l->default_value;
+
+    return ret_val;
+}
+
+
 /* print fields */
 /* returns the count of fields actually printed */
 int
@@ -855,11 +942,14 @@ print_fields(struct structure *s, struct record *r,char *buffer)
     int max_justify_len = 0;
     register int i;
     register char *d,*f;
-    char *type = s->type,justify = r->o->justify;
+    char justify = r->o->justify;
     char *indent = r->o->indent,*separator = r->o->separator;
-    char *data_start;
+    char *data_start,*rep_start,*rep_pos;
     char *field_start;
+    char *lookup_value;
     int retval = 0;
+    int replacing,just_replaced;
+    int lookup_len;
     struct print_field *pf = r->pf;
 
     if(r->o == no_output) return retval;
@@ -879,10 +969,15 @@ print_fields(struct structure *s, struct record *r,char *buffer)
         data_start = write_pos;
         pf->data = data_start;
         pf->empty = 1;
+        replacing = 0;
+        just_replaced = 0;
+        lookup_value = NULL;
+
+        if(pf->f->lookup != NULL) d = r->o->lookup;
 
         while(*d)
         {
-            if(justify != LEFT_JUSTIFY && justify != RIGHT_JUSTIFY && justify == *d && pf->justify_length == -1)
+            if(justify != LEFT_JUSTIFY && justify != RIGHT_JUSTIFY && justify == *d && pf->justify_length == -1 && !replacing)
             {
                 pf->justify_length = (int) (write_pos - data_start);
                 if(pf->justify_length > max_justify_len)
@@ -913,19 +1008,69 @@ print_fields(struct structure *s, struct record *r,char *buffer)
                         sprintf(num,"%ld",current_total_lineno);
                         writes(num);
                         break;
+                    case 'p':
+                        if(pf->f->const_data == NULL)
+                        {
+                            sprintf(num,"%d",pf->f->position + (s->type[0] == FIXED_LENGTH ? 1 : 0));
+                            writes(num);
+                        }
+                        break;
                     case '%':
                         writec('%');
                         break;
                     case 'n':
                         writes(pf->f->name);
                         break;
+                    case 'l':
+                    case 'L':
+                        if(pf->f->lookup != NULL)
+                        {
+                            if(lookup_value == NULL)
+                            {
+                                field_start = write_pos;  // misuse write buffer for temp space for field value
+                                switch(s->type[0])        // write trimmed data for search key
+                                {
+                                    case FIXED_LENGTH:
+                                        print_fixed_field('t',pf->f,buffer);
+                                        break;
+                                    case SEPARATED:
+                                        print_separated_field('t',s->quote,s->type[1],pf->f,buffer);
+                                        break;
+                                }
+                                writec(0);
+                                lookup_value = make_lookup(pf->f->lookup,field_start);
+                                write_pos = field_start;  // restore write buffer
+                            }
+                        } else if(lookup_value ==  NULL)
+                        {
+                            lookup_value = "";
+                        }
+
+                        writes(lookup_value);
+
+                        if(*d == 'L')
+                        {
+                            lookup_len = strlen(lookup_value);
+                            while(lookup_len++ < pf->f->length) writec(' ');
+                        }
+                        break;
                     case 'd':
                     case 't':
                     case 'D':
                     case 'e':
                         field_start = write_pos;
 
-                        switch(type[0])
+                        if(pf->f->rep != NULL && !replacing) // start possible replacing
+                        {
+                            replacing = 1;
+                            just_replaced = 1;
+                            rep_pos = d;
+                            rep_start = write_pos;
+                            d = pf->f->rep->value;
+                            break; // end case
+                        } 
+                        
+                        switch(s->type[0])
                         {
                             case FIXED_LENGTH:
                                 print_fixed_field(*d,pf->f,buffer);
@@ -942,8 +1087,8 @@ print_fields(struct structure *s, struct record *r,char *buffer)
                             {
                                 if(strchr(r->o->empty_chars,*f) == NULL)
                                 {
-                                     pf->empty = 0;
-                                     f = write_pos; // to stop while loop
+                                    pf->empty = 0;
+                                    f = write_pos; // to stop while loop
                                 }
                                 f++;
                             }
@@ -955,13 +1100,34 @@ print_fields(struct structure *s, struct record *r,char *buffer)
                         writec(*d);
                         break;
                 }
-                d++;
+                if(!just_replaced) d++;
             } else 
             {
                 writec(*d);
                 d++;
             }
+        
+
+            if(replacing && !*d)
+            {
+                d = rep_pos;
+                d++;
+                replacing = 0;
+                if(*rep_pos == 'D')
+                {
+                    if(write_pos - rep_start < pf->f->length)
+                    {
+                        while(write_pos - rep_start < pf->f->length) writec(' ');
+                    } else if(write_pos - rep_start > pf->f->length)
+                    {
+                        write_pos = rep_start + pf->f->length;
+                    }
+                }
+            }
+
+            if(just_replaced) just_replaced = 0;
         }
+
         if(justify == RIGHT_JUSTIFY)
         {
             pf->justify_length = (int) (write_pos - data_start);
@@ -1051,11 +1217,22 @@ make_print_list(struct include_field *fl,struct field *f)
     {
         while(fl != NULL)
         {
-            n = f;
+            n = const_field;        // first check const list
+
             while(n != NULL && (strcasecmp(fl->name,n->name) != 0))
             {
                 n = n->next;
             }
+
+            if(n == NULL)           // if not found in const list check input fields
+            {
+                n = f;
+                while(n != NULL && (strcasecmp(fl->name,n->name) != 0))
+                {
+                    n = n->next;
+                }
+            }
+
             if(n != NULL) 
             {
                 if(ret == NULL)
@@ -1069,6 +1246,7 @@ make_print_list(struct include_field *fl,struct field *f)
                 }
                 c->f = n;
                 c->next = NULL;
+                fl->found = 1;
             }
             fl = fl->next;
         }
@@ -1076,25 +1254,69 @@ make_print_list(struct include_field *fl,struct field *f)
     return ret;
 }
 
+/* check that all fields we found from current structure */
+/* return the count of unmatched fields */
+int
+check_field_list(struct include_field *fl)
+{
+    int ret = 0;
+
+    while(fl != NULL)
+    {
+        if(!fl->found) 
+        {
+            if(!fl->reported) fprintf(stderr,"%s: Field '%s' was not found in the current structure\n",program,fl->name);
+            fl->reported = 1;
+            ret++;
+        }
+        fl = fl->next;
+    }
+    return ret;
+}
+
 void
 init_structure(struct structure *s,struct record *current_record,int length, char *buffer)
 {
     struct record *r;
     struct field *f;
+    struct replace *rep;
+    struct expression *e;
+    int list_errors = 0;
 
-    if(s->header)
+    r = s->r;
+
+    while(r != NULL)
     {
-        r = s->r;
-        while(r != NULL)
+        f = r->f;
+        while(f != NULL)
         {
-            f = r->f;
-            while(f != NULL)
+            if(s->header && f->name == NULL) 
+                f->name = xstrdup(get_separated_field(f->position,s->quote,s->type,buffer));
+
+            rep = replace;
+            while(rep != NULL)
             {
-                if(f->name == NULL) f->name = xstrdup(get_separated_field(f->position,s->quote,s->type,buffer));
-                f = f->next;
+                if(strcasecmp(f->name,rep->field) == 0)
+                {
+                    f->rep = rep;
+                    rep->found = 1;
+                }
+                rep = rep->next;
             }
-            r = r->next;
+
+            e = expression;
+            while(e != NULL)
+            {
+                if(strcasecmp(f->name,e->field) == 0)
+                {
+                    e->found = 1;
+                }
+                e = e->next;
+            }
+
+            f = f->next;
         }
+        r = r->next;
     }
 
     r = s->r;
@@ -1104,12 +1326,35 @@ init_structure(struct structure *s,struct record *current_record,int length, cha
         if(r->o != no_output) r->pf = make_print_list(r->o->fl,r->f);
         r = r->next;
     }
+
+    r = s->r;
+
+    while(r != NULL)
+    {
+        if(check_field_list(r->o->fl)) list_errors++;
+        r = r->next;
+    }
+    if(list_errors) panic("Some fields from field list were not found in the current structure or constant values",s->name,NULL);
+
+    rep = replace;
+    while(rep != NULL)
+    {
+        if(!rep->found) panic("Field to be replaced was not found in the current structure",rep->field,NULL);
+        rep = rep->next;
+    }
+
+    e = expression;
+    while(e != NULL)
+    {
+        if(!e->found) panic("Field in expression was not found in the current structure",e->field,NULL);
+        e = e->next;
+    }
 }
 
 void
-invalid_input(char *file, long int lineno,int strict)
+invalid_input(char *file, long int lineno,int strict,int length)
 {
-    fprintf(stderr,"%s: Invalid input line in file \'%s\', line %ld\n",program,file,lineno);
+    fprintf(stderr,"%s: Invalid input line in file \'%s\', line %ld, line length = %d\n",program,file,lineno,length);
     if(strict) panic("Use of -l option does not cause program to abort in case of invalid input line",NULL,NULL);
 }
 
@@ -1157,6 +1402,11 @@ eval_expression(struct structure *s, int and,int invert, char *buffer)
                 case OP_NOT_EQUAL:
                     if(strcmp(write_buffer,e->value) != 0) retval++;
                     break;
+#ifdef HAVE_REGEX
+                case OP_REQEXP:
+                    if(regexec(&e->reg,write_buffer,(size_t) 0, NULL, 0) == 0) retval++;
+                    break;
+#endif
             }
         }
         if(!and && retval)
@@ -1246,29 +1496,30 @@ execute(struct structure *s,int strict, int expression_and,int expression_invert
         r = select_record(s,length,input_line);
         if(r == NULL) 
         {
-            invalid_input(current_file_name,current_file_lineno,strict);
+            invalid_input(current_file_name,current_file_lineno,strict,length);
         } else
         {
-            if(expression != NULL && (prev_record != r || prev_record == NULL))
-            {
-                init_expression_list(r);
-            }
             if(first_line)
             {
                 init_structure(s,r,length,input_line);
             }
+
+            if(expression != NULL && (prev_record != r || prev_record == NULL))
+            {
+                init_expression_list(r);
+            }
+
             if((!first_line || !headers) && r->o != no_output)
             {
                 if((r->pf == NULL && r->o->no_data == 1) || r->pf != NULL)
                 {
-                    if(r->o->header != NULL && !header_printed) 
-                    {
-                        print_header(s,r);
-                        header_printed = 1;
-                    }
                     update_field_positions(s->type,s->quote,r->f,length,input_line);
                     if(expression == NULL || (eval_expression(s,expression_and,expression_invert,input_line)))
                     {
+                        if(r->o->header != NULL && !header_printed) 
+                        {
+                            header_printed = print_header(s,r);
+                        }
                         if(r->o->indent != NULL) print_text(s,r,r->o->indent);
                         print_text(s,r,r->o->record_header);
                         fields_printed = print_fields(s,r,input_line);
