@@ -20,7 +20,7 @@
  *
  */
 
-/* $Id: ffe.c,v 1.34 2006-09-13 13:52:12 timo Exp $ */
+/* $Id: ffe.c,v 1.51 2007-05-30 07:32:48 timo Exp $ */
 
 #include "ffe.h"
 
@@ -41,10 +41,16 @@ static char *program = PACKAGE;
 static char *program = "ffe";
 #endif
 
-#ifdef VERSION
-static char *version = VERSION;
+#ifdef PACKAGE_VERSION
+static char *version = PACKAGE_VERSION;
 #else
-static char *version = "0.1.6";
+static char *version = "0.2.2";
+#endif
+
+#ifdef HOST
+static char *host = HOST;
+#else
+static char *host = "";
 #endif
 
 #ifdef PACKAGE_BUGREPORT
@@ -53,7 +59,7 @@ static char *email_address = PACKAGE_BUGREPORT;
 static char *email_address = "tjsa@iki.fi";
 #endif
 
-static char short_opts[] = "c:s:o:p:f:e:l?Vav";
+static char short_opts[] = "c:s:o:p:f:e:r:l?Vav";
 
 #ifdef HAVE_GETOPT_LONG
 static struct option long_opts[] = {
@@ -68,6 +74,7 @@ static struct option long_opts[] = {
     {"version",0,NULL,'V'},
     {"and",0,NULL,'a'},
     {"invert-match",0,NULL,'v'},
+    {"replace",1,NULL,'r'},
     {NULL,0,NULL,0}
 };
 #endif
@@ -76,6 +83,9 @@ static struct option long_opts[] = {
 struct structure *structure = NULL;
 struct output *output = NULL;
 struct expression *expression = NULL;
+struct lookup *lookup = NULL;
+struct replace *replace = NULL;
+struct field *const_field = NULL;
 
 /* output no marker */
 struct output dummy;
@@ -107,15 +117,19 @@ get_default_rc_name()
 {
     char *home;
     char *result;
+#ifdef WIN32
+    char *file = "ffe.rc";
+#else
     char *file = ".fferc";
+#endif
 
     result = NULL;
     home = getenv("HOME");
     if(home != NULL)
     {
-            result = xmalloc(strlen(home) + strlen(file) + 2);
+            result = xmalloc(strlen(home) + strlen(file) + strlen(PATH_SEPARATOR_STRING) + 2);
             strcpy(result,home);
-            strcat(result,"/");
+            strcat(result,PATH_SEPARATOR_STRING);
             strcat(result,file);
     } else
     {
@@ -138,7 +152,7 @@ help(FILE *stream)
     fprintf(stream,"-o, --output=NAME\n");
     fprintf(stream,"\t\tWrite output to NAME instead of standard output.\n");
     fprintf(stream,"-f, --field-list=LIST\n");
-    fprintf(stream,"\t\tPrint only fields listed in comma separated list LIST.\n");
+    fprintf(stream,"\t\tPrint only fields and constants listed in comma separated list LIST.\n");
     fprintf(stream,"-e, --expression=EXPRESSION\n");
     fprintf(stream,"\t\tPrint only those records for which the EXPRESSION evaluates to true.\n");
     fprintf(stream,"-a, --and\n");
@@ -147,6 +161,8 @@ help(FILE *stream)
     fprintf(stream,"\t\tPrint only those records which don't match the expression.\n");
     fprintf(stream,"-l, --loose\n");
     fprintf(stream,"\t\tAn invalid input line does not cause %s to abort.\n",program);
+    fprintf(stream,"-r, --replace=FIELD=VALUE\n");
+    fprintf(stream,"\t\tReplace FIELDs contents with VALUE in output.\n");
     fprintf(stream,"-?, --help\n");
     fprintf(stream,"\t\tDisplay this help and exit.\n");
     fprintf(stream,"-V, --version\n");
@@ -160,7 +176,7 @@ help(FILE *stream)
     fprintf(stream,"-o NAME\n");
     fprintf(stream,"\t\tWrite output to NAME instead of standard output.\n");
     fprintf(stream,"-f LIST\n");
-    fprintf(stream,"\t\tPrint only fields listed in comma separated list LIST.\n");
+    fprintf(stream,"\t\tPrint only fields and constants listed in comma separated list LIST.\n");
     fprintf(stream,"-e EXPRESSION\n");
     fprintf(stream,"\t\tPrint only those records for which the EXPRESSION evaluates to true.\n");
     fprintf(stream,"-a\n");
@@ -169,6 +185,8 @@ help(FILE *stream)
     fprintf(stream,"\t\tPrint only those records which don't match the expression.\n");
     fprintf(stream,"-l\n");
     fprintf(stream,"\t\tAn invalid input line does not cause %s to abort.\n",program);
+    fprintf(stream,"-r FIELD=VALUE\n");
+    fprintf(stream,"\t\tReplace FIELDs contents with VALUE in output.\n");
     fprintf(stream,"-?\n");
     fprintf(stream,"\t\tDisplay this help and exit.\n");
     fprintf(stream,"-V\n");
@@ -189,8 +207,8 @@ usage(int opt)
 void
 print_version()
 {
-    printf("%s version %s\n",program,version);
-    printf("Copyright (c) 2006 Timo Savinen\n\n");
+    printf("%s version %s (%s)\n",program,version,host);
+    printf("Copyright (c) 2007 Timo Savinen\n\n");
     printf("This is free software; see the source for copying conditions.\n");
     printf("There is NO warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n");
 }
@@ -249,6 +267,7 @@ check_rc(char *use_output)
     struct output *o;
     struct record *r,*fr;
     struct field *f;
+    struct lookup *l;
     int several_records;
     int errors = 0;
     int ordinal;
@@ -363,6 +382,27 @@ check_rc(char *use_output)
                     errors++;
                     fprintf(stderr,"%s: A field \'%s\' must have length in fixed length structure \'%s\' record \'%s\'\n",program,f->name,s->name,r->name);
                 }
+
+                if(f->lookup_table_name != NULL)
+                {
+                    l = lookup;
+
+                    while(l != NULL && f->lookup == NULL)
+                    {
+                        if(strcmp(l->name,f->lookup_table_name) == 0)
+                        {
+                            f->lookup = l;
+                        }
+                        l = l->next;
+                    }
+
+                    if(f->lookup == NULL)
+                    {
+                        errors++;
+                        fprintf(stderr,"%s: No lookup table named as '%s'\n",program,f->lookup_table_name);
+                    }
+                }
+
                 f = f->next;
                 ordinal++;
             }
@@ -382,20 +422,71 @@ check_rc(char *use_output)
         fprintf(stderr,"%s: No output definitions in rc-file\n",program);
     }
 
+    while(o != NULL)
+    {
+        if(o->lookup == NULL) o->lookup = o->data;
+        o = o->next;
+    }
+
     if(errors)
     {
         panic("Errors in rc-file",NULL,NULL);
     }
 }
 
+void
+add_replace(char *optarg)
+{
+    char *op_pos;
+    struct replace *r;
+
+    if((op_pos = strchr(optarg,'=')) == NULL)
+    {
+        panic("Replace expression must contain =-character",optarg,NULL);
+    }
+
+    *op_pos = 0;
+
+    r = replace;
+
+    if(r == NULL)
+    {
+        replace = xmalloc(sizeof(struct replace));
+        replace->next =  NULL;
+        r = replace;
+    } else
+    {
+        while(r->next != NULL) r = r->next;
+        r->next = xmalloc(sizeof(struct replace));
+        r = r->next;
+        r->next = NULL;
+    }
+
+    r->field = xstrdup(optarg);
+    op_pos++;
+    r->value = xstrdup(op_pos);
+    r->found = 0;
+}
+
+
 void
 add_expression(char *optarg)
 {
     char *op_pos;
     char op = 0;
     struct expression *e;
+    int rc;
+    size_t buflen;
+    char *errbuf;
 
-    if((op_pos = strchr(optarg,OP_EQUAL)) != NULL)
+    if((op_pos = strchr(optarg,OP_REQEXP)) != NULL)
+    {
+#ifdef HAVE_REGEX
+        op = OP_REQEXP;    
+#else
+        panic("Regular expressions are not supported in this system",optarg,NULL);
+#endif
+    } else if((op_pos = strchr(optarg,OP_EQUAL)) != NULL)
     {
         op = OP_EQUAL;
     } else if((op_pos = strchr(optarg,OP_START)) != NULL)
@@ -409,7 +500,7 @@ add_expression(char *optarg)
         op = OP_NOT_EQUAL;
     } else
     {
-        panic("Expression must contain an operator: =,^,~ or !",optarg,NULL);
+        panic("Expression must contain an operator: =,^,~,? or !",optarg,NULL);
     }
 
     *op_pos = 0;
@@ -433,12 +524,24 @@ add_expression(char *optarg)
     op_pos++;
     e->value = xstrdup(op_pos);
     e->value_len = strlen(e->value);
+    e->found = 0;
     e->op = op;
+#ifdef HAVE_REGEX
+    if(e->op == OP_REQEXP)
+    {
+        rc = regcomp(&e->reg,e->value,REG_EXTENDED | REG_NOSUB);
+        if(rc)
+        {
+            buflen = regerror(rc,&e->reg,NULL,0);
+            errbuf = xmalloc(buflen + 1);
+            regerror(rc,&e->reg,errbuf,buflen);
+            panic("Error in regular expression",e->value,errbuf);
+        }
+    }
+#endif
 }
        
 
-
-
 int
 main(int argc, char **argv)
 {
@@ -508,6 +611,9 @@ main(int argc, char **argv)
                 case 'e':
                     add_expression(optarg);
                     break;
+                case 'r':
+                    add_replace(optarg);
+                    break;
                 case 'a':
                     expression_and = 1;
                     break;
@@ -532,6 +638,7 @@ main(int argc, char **argv)
             }
         }
 
+
     if(optind < argc)
     {
         while(optind < argc) set_input_file(argv[optind++]);
