@@ -20,7 +20,7 @@
  *
  */
 
-/* $Id: parserc.c,v 1.24 2006-04-23 16:23:01 timo Exp $ */
+/* $Id: parserc.c,v 1.41 2007-05-30 07:32:48 timo Exp $ */
 /* parsing the rc-file */
 
 #include "ffe.h"
@@ -80,19 +80,25 @@ char *values[100];            /* poister to option name and parameters */
 #define N_RECORD            "record"
 #define N_ID                "id"
 #define N_FIELD             "field"
-#define N_FIELDSFROM         "fields-from"
-#define N_FILE_HEADER       "file_header"
-#define N_FILE_TRAILER      "file_trailer"
+#define N_FIELDSFROM        "fields-from"
+#define N_FILE_HEADER       "file-header"
+#define N_FILE_TRAILER      "file-trailer"
 #define N_DATA              "data"
 #define N_SEPARATOR         "separator"
-#define N_RECORD_HEADER     "record_header"
-#define N_RECORD_TRAILER    "record_trailer"
+#define N_RECORD_HEADER     "record-header"
+#define N_RECORD_TRAILER    "record-trailer"
 #define N_JUSTIFY           "justify"
 #define N_INDENT            "indent"
 #define N_FIELDLIST         "field-list"
 #define N_PRINT_NO_DATA     "no-data-print"
 #define N_FIELD_EMPTY_PRINT "field-empty-print"
 #define N_EMPTY_CHARS       "empty-chars"
+#define N_LOOKUP            "lookup"
+#define N_PAIR              "pair"
+#define N_FILE              "file"
+#define N_DEFAULT           "default-value"
+#define N_SEARCH            "search"
+#define N_CONST             "const"
 
 
 
@@ -105,7 +111,7 @@ static struct rc_option rc_opts[] = {
     {N_RECORD,"S"},
     {N_OUTPUT,"S"},
     {N_ID,"NS"},
-    {N_FIELD,"Sn"},
+    {N_FIELD,"Sns"},
     {N_FIELDSFROM,"S"},
     {N_FILE_HEADER,"S"},
     {N_FILE_TRAILER,"S"},
@@ -119,6 +125,12 @@ static struct rc_option rc_opts[] = {
     {N_PRINT_NO_DATA,"S"},
     {N_FIELD_EMPTY_PRINT,"S"},
     {N_EMPTY_CHARS,"S"},
+    {N_LOOKUP,"S"},
+    {N_PAIR,"SS"},
+    {N_FILE,"Sc"},
+    {N_DEFAULT,"S"},
+    {N_SEARCH,"S"},
+    {N_CONST,"SS"},
     {NULL,NULL}
 };
 
@@ -134,7 +146,7 @@ open_rc_file(char *file)
 void
 trim(char *buf)
 {
-    char *wpos=buf,*rpos=buf;
+    register char *wpos=buf,*rpos=buf;
 
     while(isspace(*rpos)) rpos++;
 
@@ -199,6 +211,8 @@ parse_include_list(char *list)
                 }
                 c->next = NULL;
                 c->name = xstrdup(s);
+                c->found = 0;
+                c->reported = 0;
             }
             p++;
             while(*p == ',') p++;
@@ -219,8 +233,8 @@ int
 read_logical_line(char *buffer, int bufsize)
 {
     static char last_eol_char = 0;      /* for what did previous read end */
-    int prev_char = 0,c;
-    int i=0,retval = 0;
+    register int prev_char = 0,c;
+    register int i=0,retval = 0;
 
     buffer[0] = 0;
 
@@ -328,6 +342,7 @@ read_logical_line(char *buffer, int bufsize)
                 retval = LL_OPTION;
                 break;
             case '\\':
+            case '\r':  /* skip win32 CR */
             case 0:
                 break;
             default:
@@ -385,7 +400,7 @@ read_logical_line(char *buffer, int bufsize)
 void 
 error_in_line()
 {
-    fprintf(stderr,"%s: Error in rc-file, line %d\n",program,lineno);
+    fprintf(stderr,"%s: Error in rcfile, line %d\n",program,lineno);
 }
 
 #define READ_BUF_SIZE 1024
@@ -395,9 +410,10 @@ error_in_line()
 /* return the number of elements in array */
 int parse_option(char *buf)
 {
-    char *rpos = buf;
+    register char *rpos = buf;
     char *end = buf;
     char *param;
+    register char *p;
     int i = 0,j;
     int valc = 0;
     int quoted;
@@ -409,6 +425,14 @@ int parse_option(char *buf)
     while(!isspace(*rpos) && *rpos) rpos++;
     *rpos = 0;
 
+    p = buf;
+
+    while(*p)  // convert _ to -
+    {
+        if(*p == '_') *p = '-';
+        p++;
+    }
+
     while(rc_opts[i].name != NULL && strcmp(rc_opts[i].name,values[0]) != 0) i++;
 
     if(rc_opts[i].name != NULL)  /* found */
@@ -488,7 +512,7 @@ int parse_option(char *buf)
                                             panic("Too many parameters",values[0],NULL);
                                         }
                                     }
-                                    if(j) *(rpos - j) = *rpos;
+                                    if(j && *rpos) *(rpos - j) = *rpos;
                                 }
                                 if(quoted)
                                 {
@@ -524,13 +548,20 @@ int parse_option(char *buf)
                         {
                             valc++;
                             values[valc] = rpos;
-                            while(isdigit(*rpos)) rpos++;
-                            if(!isspace(*rpos) && *rpos)
+                            if(*rpos == '*' && (isspace(rpos[1]) || !rpos[1]))
                             {
-                                error_in_line();
-                                panic("A number expected",values[0],NULL);
+                                rpos++;
+                                *rpos=0;
+                            } else
+                            {
+                                while(isdigit(*rpos)) rpos++;
+                                if(!isspace(*rpos) && *rpos)
+                                {
+                                    error_in_line();
+                                    panic("A number expected",values[0],NULL);
+                                }
+                                *rpos=0;
                             }
-                            *rpos=0;
                         } else
                         {
                             if(*param == 'N')
@@ -566,6 +597,105 @@ int parse_option(char *buf)
     return valc;
 }
 
+/* expand tilde in path */
+char *
+expand_home(char *path)
+{
+    char *r;
+    char *home = getenv("HOME");
+
+    if(*path == '~' && strncmp(&path[1],PATH_SEPARATOR_STRING,strlen(PATH_SEPARATOR_STRING)) == 0 && home != NULL)
+    {
+        r = xmalloc(strlen(home) + strlen(path));
+        strcpy(r,home);
+        strcat(r,&path[1]);
+    } else
+    {
+        r = xstrdup(path);
+    }
+    return r;
+}
+
+/* read key value pairs from file */
+/* to struct lookup data chain */
+/* returns pointer to last element */
+struct lookup_data *
+read_lookup_from_file(struct lookup_data **data,char *file,char separator,int *max_key_len)
+{
+    FILE *fp;
+    register int line_len;
+    size_t max_line_size = 1024;
+    char *efile;
+    char *line;
+    register char *p;
+    struct lookup_data *c_data = *data;
+
+    efile = expand_home(file);
+    
+    fp = fopen(efile,"r");
+
+    if(fp == NULL) {
+        error_in_line();
+        panic("Cannot open file",efile,strerror(errno));
+    }
+    
+    line = xmalloc(max_line_size);
+
+    if(c_data != NULL) while(c_data->next != NULL) c_data = c_data->next;
+
+    do
+    {
+#ifdef HAVE_GETLINE
+        line_len = getline(&line,&max_line_size,fp);
+#else
+        if(fgets(line,max_line_size,fp) == NULL)
+        {
+            line_len = -1;
+        } else
+        {
+            line_len = strlen(line);
+        }
+#endif
+        if(line_len > 0)
+        {
+            switch(line[line_len - 1])  // remove newline
+            {
+                case '\n':
+                case '\r':
+                    line[line_len - 1] = 0;
+                    break;
+            }
+            
+            p = line;
+            while(*p && *p != separator) p++;
+            if(*p)
+            {
+                *p = 0;
+                p++;
+
+                if(c_data == NULL) 
+                {
+                    *data = xmalloc(sizeof(struct lookup_data));
+                    c_data = *data;
+                } else
+                {
+                    c_data->next = xmalloc(sizeof(struct lookup_data));
+                    c_data = c_data->next;
+                }
+                c_data->next = NULL;
+                c_data->key = xstrdup(line);
+                c_data->value = xstrdup(p);
+                c_data->key_len = strlen(c_data->key);
+                if(*max_key_len < c_data->key_len) *max_key_len = c_data->key_len;
+            }
+        }
+    } while(line_len != -1);
+    fclose(fp);
+    free(line);
+    free(efile);
+    return c_data;
+}
+
                            
 /* parse status values */
 #define PS_MAIN 1
@@ -575,6 +705,8 @@ int parse_option(char *buf)
 #define PS_W_STRUCT 5
 #define PS_W_RECORD 6
 #define PS_W_OUTPUT 7
+#define PS_LOOKUP 8
+#define PS_W_LOOKUP 9
 
 void
 print_s()
@@ -610,6 +742,8 @@ parserc(char *rcfile,char *include_field_list)
     struct id *c_id =  NULL;
     struct record *c_record = NULL;
     struct output *c_output = output;
+    struct lookup *c_lookup = NULL;
+    struct lookup_data *c_lookup_data = NULL;
     struct include_field *fl = parse_include_list(include_field_list);
 
     char *read_buffer;
@@ -672,6 +806,7 @@ parserc(char *rcfile,char *include_field_list)
                             c_output->file_trailer = NULL;
                             c_output->header = NULL;
                             c_output->data = "%d";
+                            c_output->lookup = NULL;
                             c_output->separator = NULL;
                             c_output->record_header = NULL;
                             c_output->record_trailer = "\n";
@@ -688,10 +823,49 @@ parserc(char *rcfile,char *include_field_list)
                                 c_output->fl = NULL;
                             }
                             status = PS_W_OUTPUT;
+                        } else if(strcmp(values[0],N_LOOKUP) == 0)
+                        {
+                            if(c_lookup == NULL)
+                            {
+                                c_lookup = xmalloc(sizeof(struct lookup));
+                                lookup = c_lookup;
+                            } else
+                            {
+                                c_lookup->next = xmalloc(sizeof(struct lookup));
+                                c_lookup = c_lookup->next;
+                            }
+                            c_lookup->next = NULL;
+                            c_lookup->name = xstrdup(values[1]);
+                            c_lookup->type = EXACT;
+                            c_lookup->default_value = "";
+                            c_lookup->max_key_len = 0;
+                            c_lookup->data = NULL;
+                            status = PS_W_LOOKUP;
+                        } else if(strcmp(values[0],N_CONST) == 0)
+                        {
+                            if(const_field == NULL)
+                            {
+                                c_field = xmalloc(sizeof(struct field));
+                                const_field = c_field;
+                            } else
+                            {
+                                c_field = const_field;
+                                while(c_field->next != NULL) c_field = c_field->next;
+                                c_field->next = xmalloc(sizeof(struct field));
+                                c_field = c_field->next;
+                            }
+                            c_field->lookup_table_name = NULL;
+                            c_field->lookup = NULL;
+                            c_field->rep = NULL;
+                            c_field->next = NULL;
+                            c_field->name = xstrdup(values[1]);
+                            c_field->const_data = xstrdup(values[2]);
+                            c_field->position = 0;
+                            c_field->length = strlen(c_field->const_data);
                         } else 
                         {
                             error_in_line();
-                            panic("Option not inside structure or output",values[0],NULL);
+                            panic("Option not inside structure, output or lookup",values[0],NULL);
                         }
                         break;
                     case PS_STRUCT:
@@ -820,7 +994,12 @@ parserc(char *rcfile,char *include_field_list)
                                 c_field->next = xmalloc(sizeof(struct field));
                                 c_field = c_field->next;
                             }
+                            c_field->lookup_table_name = NULL;
+                            c_field->lookup = NULL;
+                            c_field->rep = NULL;
                             c_field->next = NULL;
+                            c_field->const_data = NULL;
+                            c_field->length = 0;
                             if(values[1][0] == '*' && !values[1][1])
                             {
                                 c_field->name = NULL;
@@ -830,15 +1009,19 @@ parserc(char *rcfile,char *include_field_list)
                             }
                             if(opt_count > 1)
                             {
-                                if(sscanf(values[2],"%d",&c_field->length) != 1)
+                                if(!(values[2][0] == '*' && !values[2][1]))
                                 {
-                                    error_in_line();
-                                    panic("Error in number",NULL,NULL);
+                                    if(sscanf(values[2],"%d",&c_field->length) != 1)
+                                    {
+                                        error_in_line();
+                                        panic("Error in number",NULL,NULL);
+                                    }
                                 }
-                            } else
-                            {
-                                c_field->length = 0;
-                            }
+                                if(opt_count > 2)
+                                {
+                                    c_field->lookup_table_name = xstrdup(values[3]);
+                                }
+                            } 
                         } else if(strcmp(values[0],N_FIELDSFROM) == 0)
                         {
                             c_record->fields_from = xstrdup(values[1]);
@@ -864,6 +1047,9 @@ parserc(char *rcfile,char *include_field_list)
                         } else if(strcmp(values[0],N_DATA) == 0)
                         {
                             c_output->data = xstrdup(values[1]);
+                        } else if (strcmp(values[0],N_LOOKUP) == 0)
+                        {
+                            c_output->lookup = xstrdup(values[1]);
                         } else if(strcmp(values[0],N_SEPARATOR) == 0)
                         {
                             c_output->separator = xstrdup(values[1]);
@@ -930,9 +1116,58 @@ parserc(char *rcfile,char *include_field_list)
                             panic("Unknown option in output definition",NULL,NULL);
                         }
                         break;
+                    case PS_LOOKUP:
+                        if(strcmp(values[0],N_SEARCH) == 0)
+                        {
+                            if(strcmp(values[1],"exact") == 0)
+                            {
+                                c_lookup->type = EXACT;
+                            } else if(strcmp(values[1],"longest") == 0)
+                            {
+                                c_lookup->type = LONGEST;
+                            } else
+                            {
+                                error_in_line();
+                                panic("Unknown value for lookup tables search option",values[1],NULL);
+                            }
+                        } else if(strcmp(values[0],N_PAIR) == 0)
+                        {
+                            if(c_lookup->data == NULL)
+                            {
+                                c_lookup->data = xmalloc(sizeof(struct lookup_data));
+                                c_lookup_data = c_lookup->data;
+                            } else
+                            {
+                                c_lookup_data->next = xmalloc(sizeof(struct lookup_data));
+                                c_lookup_data = c_lookup_data->next;
+                            }
+                            c_lookup_data->next = NULL;
+                            c_lookup_data->key = xstrdup(values[1]);
+                            c_lookup_data->value = xstrdup(values[2]);
+                            c_lookup_data->key_len = strlen(c_lookup_data->key);
+                            if(c_lookup->max_key_len < c_lookup_data->key_len) c_lookup->max_key_len = c_lookup_data->key_len;
+                        } else if(strcmp(values[0],N_FILE) == 0)
+                        {
+                            if(opt_count == 1)
+                            {
+                                c_lookup_data = read_lookup_from_file(&(c_lookup->data),values[1],';',&c_lookup->max_key_len);
+                            } else
+                            {
+                                c_lookup_data = read_lookup_from_file(&(c_lookup->data),values[1],values[2][0],&c_lookup->max_key_len);
+                            }
+                        } else if(strcmp(values[0],N_DEFAULT) == 0)
+                        {
+                            c_lookup->default_value = xstrdup(values[1]);
+                        } else 
+                        {
+                            error_in_line();
+                            panic("Unknown option for lookup",values[0],NULL);
+                        }
+                        break;
                     case PS_W_RECORD:
                     case PS_W_OUTPUT:
                     case PS_W_STRUCT:
+                    case PS_W_LOOKUP:
                         error_in_line();
                         panic("{ expected, found",values[0],NULL);
                         break;
@@ -950,6 +1185,9 @@ parserc(char *rcfile,char *include_field_list)
                     case PS_W_RECORD:
                         status = PS_RECORD;
                         break;
+                    case PS_W_LOOKUP:
+                        status = PS_LOOKUP;
+                        break;
                     default:
                         error_in_line();
                         panic("{ not expected",NULL,NULL);
@@ -961,6 +1199,7 @@ parserc(char *rcfile,char *include_field_list)
                 {
                     case PS_STRUCT:
                     case PS_OUTPUT:
+                    case PS_LOOKUP:
                         status = PS_MAIN;
                         break;
                     case PS_RECORD:
