@@ -20,11 +20,14 @@
  *
  */
 
-/* $Id: xmalloc.c,v 1.4 2006/02/13 15:53:50 timo Exp $ */
+/* $Id: xmalloc.c,v 1.6 2007-04-09 16:57:40 timo Exp $ */
 
 #include "ffe.h"
 #include <stdlib.h>
 #include <string.h>
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif
 
 void *
 xmalloc (size_t size)
